@@ -37,7 +37,7 @@ def async_signal_handler():
         for task in tasks:
             task.cancel()
         await asyncio.gather(*tasks, return_exceptions=True)
-        asyncio.get_event_loop().stop()
+        asyncio.get_running_loop().stop()
 
     asyncio.ensure_future(_handler())
 
@@ -62,7 +62,7 @@ class AsyncClient(base_client.BaseClient):
     :param http_session: an initialized ``aiohttp.ClientSession`` object to be
                          used when sending requests to the server. Use it if
                          you need to add special client options such as proxy
-                         servers, SSL certificates, etc.
+                         servers, SSL certificates, custom CA bundle, etc.
     :param ssl_verify: ``True`` to verify SSL certificates, or ``False`` to
                        skip SSL certificate verification, allowing
                        connections to servers with self signed certificates.
@@ -75,6 +75,9 @@ class AsyncClient(base_client.BaseClient):
     :param websocket_extra_options: Dictionary containing additional keyword
                                     arguments passed to
                                     ``aiohttp.ws_connect()``.
+    :param timestamp_requests: If ``True`` a timestamp is added to the query
+                               string of Socket.IO requests as a cache-busting
+                               measure. Set to ``False`` to disable.
     """
     def is_asyncio_based(self):
         return True
@@ -106,7 +109,7 @@ class AsyncClient(base_client.BaseClient):
         if self.handle_sigint and not async_signal_handler_set and \
                 threading.current_thread() == threading.main_thread():
             try:
-                asyncio.get_event_loop().add_signal_handler(
+                asyncio.get_running_loop().add_signal_handler(
                     signal.SIGINT, async_signal_handler)
             except NotImplementedError:  # pragma: no cover
                 self.logger.warning('Signal handler is unsupported')
@@ -123,7 +126,6 @@ class AsyncClient(base_client.BaseClient):
             if not transports:
                 raise ValueError('No valid transports provided')
         self.transports = transports or valid_transports
-        self.queue = self.create_queue()
         return await getattr(self, '_connect_' + self.transports[0])(
             url, headers or {}, engineio_path)
 
@@ -149,7 +151,7 @@ class AsyncClient(base_client.BaseClient):
         """
         await self._send_packet(packet.Packet(packet.MESSAGE, data=data))
 
-    async def disconnect(self, abort=False):
+    async def disconnect(self, abort=False, reason=None):
         """Disconnect from the server.
 
         :param abort: If set to ``True``, do not wait for background tasks
@@ -161,7 +163,9 @@ class AsyncClient(base_client.BaseClient):
             await self._send_packet(packet.Packet(packet.CLOSE))
             await self.queue.put(None)
             self.state = 'disconnecting'
-            await self._trigger_event('disconnect', run_async=False)
+            await self._trigger_event('disconnect',
+                                      reason or self.reason.CLIENT_DISCONNECT,
+                                      run_async=False)
             if self.current_transport == 'websocket':
                 await self.ws.close()
             if not abort:
@@ -194,11 +198,15 @@ class AsyncClient(base_client.BaseClient):
         """
         return await asyncio.sleep(seconds)
 
-    def create_queue(self):
+    def create_queue(self, *args, **kwargs):
         """Create a queue object."""
-        q = asyncio.Queue()
-        q.Empty = asyncio.QueueEmpty
-        return q
+        return asyncio.Queue(*args, **kwargs)
+
+    def get_queue_empty_exception(self):
+        """Return the queue empty exception raised by queues created by the
+        ``create_queue()`` method.
+        """
+        return asyncio.QueueEmpty
 
     def create_event(self):
         """Create an event object."""
@@ -409,7 +417,8 @@ class AsyncClient(base_client.BaseClient):
         elif pkt.packet_type == packet.PING:
             await self._send_packet(packet.Packet(packet.PONG, pkt.data))
         elif pkt.packet_type == packet.CLOSE:
-            await self.disconnect(abort=True)
+            await self.disconnect(abort=True,
+                                  reason=self.reason.SERVER_DISCONNECT)
         elif pkt.packet_type == packet.NOOP:
             pass
         else:
@@ -462,7 +471,16 @@ class AsyncClient(base_client.BaseClient):
                     return task
                 else:
                     try:
-                        ret = await self.handlers[event](*args)
+                        try:
+                            ret = await self.handlers[event](*args)
+                        except TypeError:
+                            if event == 'disconnect' and \
+                                    len(args) == 1:  # pragma: no branch
+                                # legacy disconnect events do not have a reason
+                                # argument
+                                return await self.handlers[event]()
+                            else:  # pragma: no cover
+                                raise
                     except asyncio.CancelledError:  # pragma: no cover
                         pass
                     except:
@@ -482,7 +500,16 @@ class AsyncClient(base_client.BaseClient):
                     return task
                 else:
                     try:
-                        ret = self.handlers[event](*args)
+                        try:
+                            ret = self.handlers[event](*args)
+                        except TypeError:
+                            if event == 'disconnect' and \
+                                    len(args) == 1:  # pragma: no branch
+                                # legacy disconnect events do not have a reason
+                                # argument
+                                ret = self.handlers[event]()
+                            else:  # pragma: no cover
+                                raise
                     except:
                         self.logger.exception(event + ' handler error')
                         if event == 'connect':
@@ -493,7 +520,7 @@ class AsyncClient(base_client.BaseClient):
 
     async def _read_loop_polling(self):
         """Read packets by polling the Engine.IO server."""
-        while self.state == 'connected':
+        while self.state == 'connected' and self.write_loop_task:
             self.logger.info(
                 'Sending polling GET request to ' + self.base_url)
             r = await self._send_request(
@@ -520,10 +547,12 @@ class AsyncClient(base_client.BaseClient):
             for pkt in p.packets:
                 await self._receive_packet(pkt)
 
-        self.logger.info('Waiting for write loop task to end')
-        await self.write_loop_task
+        if self.write_loop_task:  # pragma: no branch
+            self.logger.info('Waiting for write loop task to end')
+            await self.write_loop_task
         if self.state == 'connected':
-            await self._trigger_event('disconnect', run_async=False)
+            await self._trigger_event(
+                'disconnect', self.reason.TRANSPORT_ERROR, run_async=False)
             try:
                 base_client.connected_clients.remove(self)
             except ValueError:  # pragma: no cover
@@ -541,8 +570,10 @@ class AsyncClient(base_client.BaseClient):
                     timeout=self.ping_interval + self.ping_timeout)
                 if not isinstance(p.data, (str, bytes)):  # pragma: no cover
                     self.logger.warning(
-                        'Server sent unexpected packet %s data %s, aborting',
-                        str(p.type), str(p.data))
+                        'Server sent %s packet data %s, aborting',
+                        'close' if p.type in [aiohttp.WSMsgType.CLOSE,
+                                              aiohttp.WSMsgType.CLOSING]
+                        else str(p.type), str(p.data))
                     await self.queue.put(None)
                     break  # the connection is broken
                 p = p.data
@@ -571,10 +602,12 @@ class AsyncClient(base_client.BaseClient):
                 break
             await self._receive_packet(pkt)
 
-        self.logger.info('Waiting for write loop task to end')
-        await self.write_loop_task
+        if self.write_loop_task:  # pragma: no branch
+            self.logger.info('Waiting for write loop task to end')
+            await self.write_loop_task
         if self.state == 'connected':
-            await self._trigger_event('disconnect', run_async=False)
+            await self._trigger_event(
+                'disconnect', self.reason.TRANSPORT_ERROR, run_async=False)
             try:
                 base_client.connected_clients.remove(self)
             except ValueError:  # pragma: no cover
@@ -594,7 +627,7 @@ class AsyncClient(base_client.BaseClient):
             packets = None
             try:
                 packets = [await asyncio.wait_for(self.queue.get(), timeout)]
-            except (self.queue.Empty, asyncio.TimeoutError):
+            except (self.queue_empty, asyncio.TimeoutError):
                 self.logger.error('packet queue is empty, aborting')
                 break
             except asyncio.CancelledError:  # pragma: no cover
@@ -606,7 +639,7 @@ class AsyncClient(base_client.BaseClient):
                 while True:
                     try:
                         packets.append(self.queue.get_nowait())
-                    except self.queue.Empty:
+                    except self.queue_empty:
                         break
                     if packets[-1] is None:
                         packets = packets[:-1]
@@ -630,7 +663,7 @@ class AsyncClient(base_client.BaseClient):
                 if r.status < 200 or r.status >= 300:
                     self.logger.warning('Unexpected status code %s in server '
                                         'response, aborting', r.status)
-                    await self._reset()
+                    self.write_loop_task = None
                     break
             else:
                 # websocket
