@@ -1,15 +1,12 @@
+from base64 import b64encode
+from engineio.json import JSONDecodeError
 import logging
-try:
-    import queue
-except ImportError:  # pragma: no cover
-    import Queue as queue
-import signal
+import queue
 import ssl
 import threading
 import time
+import urllib
 
-import six
-from six.moves import urllib
 try:
     import requests
 except ImportError:  # pragma: no cover
@@ -18,36 +15,15 @@ try:
     import websocket
 except ImportError:  # pragma: no cover
     websocket = None
+from . import base_client
 from . import exceptions
 from . import packet
 from . import payload
 
 default_logger = logging.getLogger('engineio.client')
-connected_clients = []
-
-if six.PY2:  # pragma: no cover
-    ConnectionError = OSError
 
 
-def signal_handler(sig, frame):
-    """SIGINT handler.
-
-    Disconnect all active clients and then invoke the original signal handler.
-    """
-    for client in connected_clients[:]:
-        if not client.is_asyncio_based():
-            client.disconnect()
-    if callable(original_signal_handler):
-        return original_signal_handler(sig, frame)
-    else:  # pragma: no cover
-        # Handle case where no original SIGINT handler was present.
-        return signal.default_int_handler(sig, frame)
-
-
-original_signal_handler = None
-
-
-class Client(object):
+class Client(base_client.BaseClient):
     """An Engine.IO client.
 
     This class implements a fully compliant Engine.IO web client with support
@@ -63,95 +39,27 @@ class Client(object):
                  versions.
     :param request_timeout: A timeout in seconds for requests. The default is
                             5 seconds.
+    :param http_session: an initialized ``requests.Session`` object to be used
+                         when sending requests to the server. Use it if you
+                         need to add special client options such as proxy
+                         servers, SSL certificates, custom CA bundle, etc.
     :param ssl_verify: ``True`` to verify SSL certificates, or ``False`` to
                        skip SSL certificate verification, allowing
                        connections to servers with self signed certificates.
                        The default is ``True``.
+    :param handle_sigint: Set to ``True`` to automatically handle disconnection
+                          when the process is interrupted, or to ``False`` to
+                          leave interrupt handling to the calling application.
+                          Interrupt handling can only be enabled when the
+                          client instance is created in the main thread.
+    :param websocket_extra_options: Dictionary containing additional keyword
+                                    arguments passed to
+                                    ``websocket.create_connection()``.
+    :param timestamp_requests: If ``True`` a timestamp is added to the query
+                               string of Socket.IO requests as a cache-busting
+                               measure. Set to ``False`` to disable.
     """
-    event_names = ['connect', 'disconnect', 'message']
-
-    def __init__(self,
-                 logger=False,
-                 json=None,
-                 request_timeout=5,
-                 ssl_verify=True):
-        global original_signal_handler
-        if original_signal_handler is None and \
-                threading.current_thread() == threading.main_thread():
-            original_signal_handler = signal.signal(signal.SIGINT,
-                                                    signal_handler)
-        self.handlers = {}
-        self.base_url = None
-        self.transports = None
-        self.current_transport = None
-        self.sid = None
-        self.upgrades = None
-        self.ping_interval = None
-        self.ping_timeout = None
-        self.pong_received = True
-        self.http = None
-        self.ws = None
-        self.read_loop_task = None
-        self.write_loop_task = None
-        self.ping_loop_task = None
-        self.ping_loop_event = None
-        self.queue = None
-        self.state = 'disconnected'
-        self.ssl_verify = ssl_verify
-
-        if json is not None:
-            packet.Packet.json = json
-        if not isinstance(logger, bool):
-            self.logger = logger
-        else:
-            self.logger = default_logger
-            if not logging.root.handlers and \
-                    self.logger.level == logging.NOTSET:
-                if logger:
-                    self.logger.setLevel(logging.INFO)
-                else:
-                    self.logger.setLevel(logging.ERROR)
-                self.logger.addHandler(logging.StreamHandler())
-
-        self.request_timeout = request_timeout
-
-    def is_asyncio_based(self):
-        return False
-
-    def on(self, event, handler=None):
-        """Register an event handler.
-
-        :param event: The event name. Can be ``'connect'``, ``'message'`` or
-                      ``'disconnect'``.
-        :param handler: The function that should be invoked to handle the
-                        event. When this parameter is not given, the method
-                        acts as a decorator for the handler function.
-
-        Example usage::
-
-            # as a decorator:
-            @eio.on('connect')
-            def connect_handler():
-                print('Connection request')
-
-            # as a method:
-            def message_handler(msg):
-                print('Received message: ', msg)
-                eio.send('response')
-            eio.on('message', message_handler)
-        """
-        if event not in self.event_names:
-            raise ValueError('Invalid event')
-
-        def set_handler(handler):
-            self.handlers[event] = handler
-            return handler
-
-        if handler is None:
-            return set_handler
-        set_handler(handler)
-
-    def connect(self, url, headers={}, transports=None,
+    def connect(self, url, headers=None, transports=None,
                 engineio_path='engine.io'):
         """Connect to an Engine.IO server.
 
@@ -176,16 +84,15 @@ class Client(object):
             raise ValueError('Client is not in a disconnected state')
         valid_transports = ['polling', 'websocket']
         if transports is not None:
-            if isinstance(transports, six.string_types):
+            if isinstance(transports, str):
                 transports = [transports]
             transports = [transport for transport in transports
                           if transport in valid_transports]
             if not transports:
                 raise ValueError('No valid transports provided')
         self.transports = transports or valid_transports
-        self.queue = self.create_queue()
         return getattr(self, '_connect_' + self.transports[0])(
-            url, headers, engineio_path)
+            url, headers or {}, engineio_path)
 
     def wait(self):
         """Wait until the connection with the server ends.
@@ -196,21 +103,16 @@ class Client(object):
         if self.read_loop_task:
             self.read_loop_task.join()
 
-    def send(self, data, binary=None):
-        """Send a message to a client.
+    def send(self, data):
+        """Send a message to the server.
 
-        :param data: The data to send to the client. Data can be of type
+        :param data: The data to send to the server. Data can be of type
                      ``str``, ``bytes``, ``list`` or ``dict``. If a ``list``
                      or ``dict``, the data will be serialized as JSON.
-        :param binary: ``True`` to send packet as binary, ``False`` to send
-                       as text. If not given, unicode (Python 2) and str
-                       (Python 3) are sent as text, and str (Python 2) and
-                       bytes (Python 3) are sent as binary.
         """
-        self._send_packet(packet.Packet(packet.MESSAGE, data=data,
-                                        binary=binary))
+        self._send_packet(packet.Packet(packet.MESSAGE, data=data))
 
-    def disconnect(self, abort=False):
+    def disconnect(self, abort=False, reason=None):
         """Disconnect from the server.
 
         :param abort: If set to ``True``, do not wait for background tasks
@@ -220,26 +122,20 @@ class Client(object):
             self._send_packet(packet.Packet(packet.CLOSE))
             self.queue.put(None)
             self.state = 'disconnecting'
-            self._trigger_event('disconnect', run_async=False)
+            self._trigger_event('disconnect',
+                                reason or self.reason.CLIENT_DISCONNECT,
+                                run_async=False)
             if self.current_transport == 'websocket':
                 self.ws.close()
             if not abort:
                 self.read_loop_task.join()
             self.state = 'disconnected'
             try:
-                connected_clients.remove(self)
+                base_client.connected_clients.remove(self)
             except ValueError:  # pragma: no cover
                 pass
         self._reset()
 
-    def transport(self):
-        """Return the name of the transport currently in use.
-
-        The possible values returned by this function are ``'polling'`` and
-        ``'websocket'``.
-        """
-        return self.current_transport
-
     def start_background_task(self, target, *args, **kwargs):
         """Start a background task.
 
@@ -250,11 +146,12 @@ class Client(object):
         :param args: arguments to pass to the function.
         :param kwargs: keyword arguments to pass to the function.
 
-        This function returns an object compatible with the `Thread` class in
-        the Python standard library. The `start()` method on this object is
-        already called by this function.
+        This function returns an object that represents the background task,
+        on which the ``join()`` method can be invoked to wait for the task to
+        complete.
         """
-        th = threading.Thread(target=target, args=args, kwargs=kwargs)
+        th = threading.Thread(target=target, args=args, kwargs=kwargs,
+                              daemon=True)
         th.start()
         return th
 
@@ -264,18 +161,18 @@ class Client(object):
 
     def create_queue(self, *args, **kwargs):
         """Create a queue object."""
-        q = queue.Queue(*args, **kwargs)
-        q.Empty = queue.Empty
-        return q
+        return queue.Queue(*args, **kwargs)
+
+    def get_queue_empty_exception(self):
+        """Return the queue empty exception raised by queues created by the
+        ``create_queue()`` method.
+        """
+        return queue.Empty
 
     def create_event(self, *args, **kwargs):
         """Create an event object."""
         return threading.Event(*args, **kwargs)
 
-    def _reset(self):
-        self.state = 'disconnected'
-        self.sid = None
-
     def _connect_polling(self, url, headers, engineio_path):
         """Establish a long-polling connection to the Engine.IO server."""
         if requests is None:  # pragma: no cover
@@ -288,20 +185,24 @@ class Client(object):
         r = self._send_request(
             'GET', self.base_url + self._get_url_timestamp(), headers=headers,
             timeout=self.request_timeout)
-        if r is None:
+        if r is None or isinstance(r, str):
             self._reset()
             raise exceptions.ConnectionError(
-                'Connection refused by the server')
+                r or 'Connection refused by the server')
         if r.status_code < 200 or r.status_code >= 300:
             self._reset()
+            try:
+                arg = r.json()
+            except JSONDecodeError:
+                arg = None
             raise exceptions.ConnectionError(
                 'Unexpected status code {} in server response'.format(
-                    r.status_code), r.json())
+                    r.status_code), arg)
         try:
-            p = payload.Payload(encoded_payload=r.content)
+            p = payload.Payload(encoded_payload=r.content.decode('utf-8'))
         except ValueError:
-            six.raise_from(exceptions.ConnectionError(
-                'Unexpected response from server'), None)
+            raise exceptions.ConnectionError(
+                'Unexpected response from server') from None
         open_packet = p.packets[0]
         if open_packet.packet_type != packet.OPEN:
             raise exceptions.ConnectionError(
@@ -310,13 +211,13 @@ class Client(object):
             'Polling connection accepted with ' + str(open_packet.data))
         self.sid = open_packet.data['sid']
         self.upgrades = open_packet.data['upgrades']
-        self.ping_interval = open_packet.data['pingInterval'] / 1000.0
-        self.ping_timeout = open_packet.data['pingTimeout'] / 1000.0
+        self.ping_interval = int(open_packet.data['pingInterval']) / 1000.0
+        self.ping_timeout = int(open_packet.data['pingTimeout']) / 1000.0
         self.current_transport = 'polling'
         self.base_url += '&sid=' + self.sid
 
         self.state = 'connected'
-        connected_clients.append(self)
+        base_client.connected_clients.append(self)
         self._trigger_event('connect', run_async=False)
 
         for pkt in p.packets[1:]:
@@ -329,7 +230,6 @@ class Client(object):
                 return
 
         # start background tasks associated with this client
-        self.ping_loop_task = self.start_background_task(self._ping_loop)
         self.write_loop_task = self.start_background_task(self._write_loop)
         self.read_loop_task = self.start_background_task(
             self._read_loop_polling)
@@ -338,8 +238,8 @@ class Client(object):
         """Establish or upgrade to a WebSocket connection with the server."""
         if websocket is None:  # pragma: no cover
             # not installed
-            self.logger.warning('websocket-client package not installed, only '
-                                'polling transport is available')
+            self.logger.error('websocket-client package not installed, only '
+                              'polling transport is available')
             return False
         websocket_url = self._get_engineio_url(url, engineio_path, 'websocket')
         if self.sid:
@@ -353,11 +253,13 @@ class Client(object):
             self.logger.info(
                 'Attempting WebSocket connection to ' + websocket_url)
 
-        # get the cookies from the long-polling connection so that they can
-        # also be sent the the WebSocket route
+        # get cookies and other settings from the long-polling connection
+        # so that they are preserved when connecting to the WebSocket route
         cookies = None
+        extra_options = {}
         if self.http:
-            cookies = '; '.join(["{}={}".format(cookie.name, cookie.value)
+            # cookies
+            cookies = '; '.join([f"{cookie.name}={cookie.value}"
                                  for cookie in self.http.cookies])
             for header, value in headers.items():
                 if header.lower() == 'cookie':
@@ -367,14 +269,71 @@ class Client(object):
                     del headers[header]
                     break
 
-        extra_options = {}
+            # auth
+            if 'Authorization' not in headers and self.http.auth is not None:
+                if not isinstance(self.http.auth, tuple):  # pragma: no cover
+                    raise ValueError('Only basic authentication is supported')
+                basic_auth = '{}:{}'.format(
+                    self.http.auth[0], self.http.auth[1]).encode('utf-8')
+                basic_auth = b64encode(basic_auth).decode('utf-8')
+                headers['Authorization'] = 'Basic ' + basic_auth
+
+            # cert
+            # this can be given as ('certfile', 'keyfile') or just 'certfile'
+            if isinstance(self.http.cert, tuple):
+                extra_options['sslopt'] = {
+                    'certfile': self.http.cert[0],
+                    'keyfile': self.http.cert[1]}
+            elif self.http.cert:
+                extra_options['sslopt'] = {'certfile': self.http.cert}
+
+            # proxies
+            if self.http.proxies:
+                proxy_url = None
+                if websocket_url.startswith('ws://'):
+                    proxy_url = self.http.proxies.get(
+                        'ws', self.http.proxies.get('http'))
+                else:  # wss://
+                    proxy_url = self.http.proxies.get(
+                        'wss', self.http.proxies.get('https'))
+                if proxy_url:
+                    parsed_url = urllib.parse.urlparse(
+                        proxy_url if '://' in proxy_url
+                        else 'scheme://' + proxy_url)
+                    extra_options['http_proxy_host'] = parsed_url.hostname
+                    extra_options['http_proxy_port'] = parsed_url.port
+                    extra_options['http_proxy_auth'] = (
+                        (parsed_url.username, parsed_url.password)
+                        if parsed_url.username or parsed_url.password
+                        else None)
+
+            # verify
+            if isinstance(self.http.verify, str):
+                if 'sslopt' in extra_options:
+                    extra_options['sslopt']['ca_certs'] = self.http.verify
+                else:
+                    extra_options['sslopt'] = {'ca_certs': self.http.verify}
+            elif not self.http.verify:
+                self.ssl_verify = False
+
         if not self.ssl_verify:
-            extra_options['sslopt'] = {"cert_reqs": ssl.CERT_NONE}
+            if 'sslopt' in extra_options:
+                extra_options['sslopt'].update({"cert_reqs": ssl.CERT_NONE})
+            else:
+                extra_options['sslopt'] = {"cert_reqs": ssl.CERT_NONE}
+
+        # combine internally generated options with the ones supplied by the
+        # caller. The caller's options take precedence.
+        headers.update(self.websocket_extra_options.pop('header', {}))
+        extra_options['header'] = headers
+        extra_options['cookie'] = cookies
+        extra_options['enable_multithread'] = True
+        extra_options['timeout'] = self.request_timeout
+        extra_options.update(self.websocket_extra_options)
         try:
             ws = websocket.create_connection(
-                websocket_url + self._get_url_timestamp(), header=headers,
-                cookie=cookies, enable_multithread=True, **extra_options)
-        except (ConnectionError, IOError, websocket.WebSocketException):
+                websocket_url + self._get_url_timestamp(), **extra_options)
+        except (ConnectionError, OSError, websocket.WebSocketException):
             if upgrade:
                 self.logger.warning(
                     'WebSocket upgrade failed: connection error')
@@ -382,8 +341,7 @@ class Client(object):
             else:
                 raise exceptions.ConnectionError('Connection error')
         if upgrade:
-            p = packet.Packet(packet.PING,
-                              data=six.text_type('probe')).encode()
+            p = packet.Packet(packet.PING, data='probe').encode()
             try:
                 ws.send(p)
             except Exception as e:  # pragma: no cover
@@ -426,17 +384,17 @@ class Client(object):
                 'WebSocket connection accepted with ' + str(open_packet.data))
             self.sid = open_packet.data['sid']
             self.upgrades = open_packet.data['upgrades']
-            self.ping_interval = open_packet.data['pingInterval'] / 1000.0
-            self.ping_timeout = open_packet.data['pingTimeout'] / 1000.0
+            self.ping_interval = int(open_packet.data['pingInterval']) / 1000.0
+            self.ping_timeout = int(open_packet.data['pingTimeout']) / 1000.0
             self.current_transport = 'websocket'
 
             self.state = 'connected'
-            connected_clients.append(self)
+            base_client.connected_clients.append(self)
             self._trigger_event('connect', run_async=False)
         self.ws = ws
+        self.ws.settimeout(self.ping_interval + self.ping_timeout)
 
         # start background tasks associated with this client
-        self.ping_loop_task = self.start_background_task(self._ping_loop)
         self.write_loop_task = self.start_background_task(self._write_loop)
         self.read_loop_task = self.start_background_task(
             self._read_loop_websocket)
@@ -451,10 +409,10 @@ class Client(object):
             pkt.data if not isinstance(pkt.data, bytes) else '<binary>')
         if pkt.packet_type == packet.MESSAGE:
             self._trigger_event('message', pkt.data, run_async=True)
-        elif pkt.packet_type == packet.PONG:
-            self.pong_received = True
+        elif pkt.packet_type == packet.PING:
+            self._send_packet(packet.Packet(packet.PONG, pkt.data))
         elif pkt.packet_type == packet.CLOSE:
-            self.disconnect(abort=True)
+            self.disconnect(abort=True, reason=self.reason.SERVER_DISCONNECT)
         elif pkt.packet_type == packet.NOOP:
             pass
         else:
@@ -476,12 +434,15 @@ class Client(object):
             timeout=None):  # pragma: no cover
         if self.http is None:
             self.http = requests.Session()
+        if not self.ssl_verify:
+            self.http.verify = False
         try:
             return self.http.request(method, url, headers=headers, data=body,
-                                     timeout=timeout, verify=self.ssl_verify)
+                                     timeout=timeout)
         except requests.exceptions.RequestException as exc:
             self.logger.info('HTTP %s request to %s failed with error %s.',
                              method, url, exc)
+            return str(exc)
 
     def _trigger_event(self, event, *args, **kwargs):
         """Invoke an event handler."""
@@ -491,68 +452,30 @@ class Client(object):
                 return self.start_background_task(self.handlers[event], *args)
             else:
                 try:
-                    return self.handlers[event](*args)
+                    try:
+                        return self.handlers[event](*args)
+                    except TypeError:
+                        if event == 'disconnect' and \
+                                len(args) == 1:  # pragma: no branch
+                            # legacy disconnect events do  not have a reason
+                            # argument
+                            return self.handlers[event]()
+                        else:  # pragma: no cover
+                            raise
                 except:
                     self.logger.exception(event + ' handler error')
 
-    def _get_engineio_url(self, url, engineio_path, transport):
-        """Generate the Engine.IO connection URL."""
-        engineio_path = engineio_path.strip('/')
-        parsed_url = urllib.parse.urlparse(url)
-
-        if transport == 'polling':
-            scheme = 'http'
-        elif transport == 'websocket':
-            scheme = 'ws'
-        else:  # pragma: no cover
-            raise ValueError('invalid transport')
-        if parsed_url.scheme in ['https', 'wss']:
-            scheme += 's'
-
-        return ('{scheme}://{netloc}/{path}/?{query}'
-                '{sep}transport={transport}&EIO=3').format(
-                    scheme=scheme, netloc=parsed_url.netloc,
-                    path=engineio_path, query=parsed_url.query,
-                    sep='&' if parsed_url.query else '',
-                    transport=transport)
-
-    def _get_url_timestamp(self):
-        """Generate the Engine.IO query string timestamp."""
-        return '&t=' + str(time.time())
-
-    def _ping_loop(self):
-        """This background task sends a PING to the server at the requested
-        interval.
-        """
-        self.pong_received = True
-        if self.ping_loop_event is None:
-            self.ping_loop_event = self.create_event()
-        else:
-            self.ping_loop_event.clear()
-        while self.state == 'connected':
-            if not self.pong_received:
-                self.logger.info(
-                    'PONG response has not been received, aborting')
-                if self.ws:
-                    self.ws.close(timeout=0)
-                self.queue.put(None)
-                break
-            self.pong_received = False
-            self._send_packet(packet.Packet(packet.PING))
-            self.ping_loop_event.wait(timeout=self.ping_interval)
-        self.logger.info('Exiting ping task')
-
     def _read_loop_polling(self):
         """Read packets by polling the Engine.IO server."""
-        while self.state == 'connected':
+        while self.state == 'connected' and self.write_loop_task:
             self.logger.info(
                 'Sending polling GET request to ' + self.base_url)
             r = self._send_request(
                 'GET', self.base_url + self._get_url_timestamp(),
                 timeout=max(self.ping_interval, self.ping_timeout) + 5)
-            if r is None:
+            if r is None or isinstance(r, str):
                 self.logger.warning(
-                    'Connection refused by the server, aborting')
+                    r or 'Connection refused by the server, aborting')
                 self.queue.put(None)
                 break
             if r.status_code < 200 or r.status_code >= 300:
@@ -561,7 +484,7 @@ class Client(object):
                 self.queue.put(None)
                 break
             try:
-                p = payload.Payload(encoded_payload=r.content)
+                p = payload.Payload(encoded_payload=r.content.decode('utf-8'))
             except ValueError:
                 self.logger.warning(
                     'Unexpected packet from server, aborting')
@@ -570,16 +493,14 @@ class Client(object):
             for pkt in p.packets:
                 self._receive_packet(pkt)
 
-        self.logger.info('Waiting for write loop task to end')
-        self.write_loop_task.join()
-        self.logger.info('Waiting for ping loop task to end')
-        if self.ping_loop_event:  # pragma: no cover
-            self.ping_loop_event.set()
-        self.ping_loop_task.join()
+        if self.write_loop_task:  # pragma: no branch
+            self.logger.info('Waiting for write loop task to end')
+            self.write_loop_task.join()
         if self.state == 'connected':
-            self._trigger_event('disconnect', run_async=False)
+            self._trigger_event('disconnect', self.reason.TRANSPORT_ERROR,
+                                run_async=False)
             try:
-                connected_clients.remove(self)
+                base_client.connected_clients.remove(self)
             except ValueError:  # pragma: no cover
                 pass
             self._reset()
@@ -591,19 +512,29 @@ class Client(object):
             p = None
             try:
                 p = self.ws.recv()
+                if len(p) == 0 and not self.ws.connected:  # pragma: no cover
+                    # websocket client can return an empty string after close
+                    raise websocket.WebSocketConnectionClosedException()
+            except websocket.WebSocketTimeoutException:
+                self.logger.warning(
+                    'Server has stopped communicating, aborting')
+                self.queue.put(None)
+                break
             except websocket.WebSocketConnectionClosedException:
                 self.logger.warning(
                     'WebSocket connection was closed, aborting')
                 self.queue.put(None)
                 break
-            except Exception as e:
-                self.logger.info(
-                    'Unexpected error receiving packet: "%s", aborting',
-                    str(e))
+            except Exception as e:  # pragma: no cover
+                if type(e) is OSError and e.errno == 9:
+                    self.logger.info(
+                        'WebSocket connection is closing, aborting')
+                else:
+                    self.logger.info(
+                        'Unexpected error receiving packet: "%s", aborting',
+                        str(e))
                 self.queue.put(None)
                 break
-            if isinstance(p, six.text_type):  # pragma: no cover
-                p = p.encode('utf-8')
             try:
                 pkt = packet.Packet(encoded_packet=p)
             except Exception as e:  # pragma: no cover
@@ -613,16 +544,14 @@ class Client(object):
                 break
             self._receive_packet(pkt)
 
-        self.logger.info('Waiting for write loop task to end')
-        self.write_loop_task.join()
-        self.logger.info('Waiting for ping loop task to end')
-        if self.ping_loop_event:  # pragma: no cover
-            self.ping_loop_event.set()
-        self.ping_loop_task.join()
+        if self.write_loop_task:  # pragma: no branch
+            self.logger.info('Waiting for write loop task to end')
+            self.write_loop_task.join()
         if self.state == 'connected':
-            self._trigger_event('disconnect', run_async=False)
+            self._trigger_event('disconnect', self.reason.TRANSPORT_ERROR,
+                                run_async=False)
             try:
-                connected_clients.remove(self)
+                base_client.connected_clients.remove(self)
             except ValueError:  # pragma: no cover
                 pass
             self._reset()
@@ -640,7 +569,7 @@ class Client(object):
             packets = None
             try:
                 packets = [self.queue.get(timeout=timeout)]
-            except self.queue.Empty:
+            except self.queue_empty:
                 self.logger.error('packet queue is empty, aborting')
                 break
             if packets == [None]:
@@ -650,7 +579,7 @@ class Client(object):
                 while True:
                     try:
                         packets.append(self.queue.get(block=False))
-                    except self.queue.Empty:
+                    except self.queue_empty:
                         break
                     if packets[-1] is None:
                         packets = packets[:-1]
@@ -663,30 +592,31 @@ class Client(object):
                 p = payload.Payload(packets=packets)
                 r = self._send_request(
                     'POST', self.base_url, body=p.encode(),
-                    headers={'Content-Type': 'application/octet-stream'},
+                    headers={'Content-Type': 'text/plain'},
                     timeout=self.request_timeout)
                 for pkt in packets:
                     self.queue.task_done()
-                if r is None:
+                if r is None or isinstance(r, str):
                     self.logger.warning(
-                        'Connection refused by the server, aborting')
+                        r or 'Connection refused by the server, aborting')
                     break
                 if r.status_code < 200 or r.status_code >= 300:
                     self.logger.warning('Unexpected status code %s in server '
                                         'response, aborting', r.status_code)
-                    self._reset()
+                    self.write_loop_task = None
                     break
             else:
                 # websocket
                 try:
                     for pkt in packets:
-                        encoded_packet = pkt.encode(always_bytes=False)
+                        encoded_packet = pkt.encode()
                         if pkt.binary:
                             self.ws.send_binary(encoded_packet)
                         else:
                             self.ws.send(encoded_packet)
                         self.queue.task_done()
-                except websocket.WebSocketConnectionClosedException:
+                except (websocket.WebSocketConnectionClosedException,
+                        BrokenPipeError, OSError):
                     self.logger.warning(
                         'WebSocket connection was closed, aborting')
                     break
